@@ -716,14 +716,14 @@ expand_cplus_start_catch (raise_id)
     }
   expand_start_cond (cond, 0);
   /* Does nothing right now.  */
-  expand_start_catch (decl, 0, 0);
+  expand_catch (decl);
   if (current_exception_type
       && TYPE_NEEDS_DESTRUCTOR (current_exception_type))
     {
       /* Make a cleanup for the name-specific exception object now in scope.  */
       tree cleanup = maybe_build_cleanup (current_exception_object);
       expand_start_bindings (0);
-      expand_decl (NULL_TREE, cleanup);
+      expand_decl_cleanup (NULL_TREE, cleanup);
     }
 }
 
@@ -736,7 +736,7 @@ expand_cplus_end_catch ()
       /* Destroy the specific exception object now in scope.  */
       expand_end_bindings (getdecls (), 1, 1);
     }
-  if (! expand_end_catch ())
+  if (! expand_escape_except ())
     abort ();
   expand_end_cond ();
 }
@@ -794,10 +794,11 @@ setup_exception_throw_decl ()
 		     build_tree_list (0, integer_one_node));
   call_to_longjmp = build_function_call (BILJ, parms);
 
-  expand_decl (exception_throw_decl,
-	       build (COND_EXPR, void_type_node,
-		      exception_throw_decl,
-		      call_to_longjmp, integer_zero_node));
+  expand_decl (exception_throw_decl);
+  expand_decl_cleanup (exception_throw_decl,
+		       build (COND_EXPR, void_type_node,
+			      exception_throw_decl,
+			      call_to_longjmp, integer_zero_node));
   DECL_INITIAL (exception_throw_decl) = integer_zero_node;
   sets_exception_throw_decl = 0;
   resume_momentary (old);
