@@ -478,8 +478,8 @@ get_base_type (parent, type, protect)
 
     while (tp < search_tail)
       {
-	CLASSTYPE_MARKED5 (*tp++) = 0;
-	tp += 1;
+	CLASSTYPE_MARKED5 (*tp) = 0;
+	tp += 2;
       }
   }
   search_stack = pop_search_level (search_stack);
@@ -508,6 +508,10 @@ get_base_type (parent, type, protect)
 }
 
 /* Return the number of levels between type PARENT and type TYPE,
+   following the leftmost path to PARENT.  If PARENT
+   appears in different places from TYPE's point of view,
+   the leftmost PARENT will be the one chosen.
+
    Return -1 if TYPE is not derived from PARENT.
    Return -2 if PARENT is an ambiguous base class of TYPE.
    Return -3 if PARENT is private to TYPE, and protect is non-zero.
@@ -625,14 +629,92 @@ get_base_distance (parent, type, protect, path_ptr)
   {
     tree *tp = search_stack->first;
     tree *search_tail = tp + tail;
+    int increment = path_ptr ? 4 : 3;
 
     while (tp < search_tail)
       {
-	CLASSTYPE_MARKED5 (*tp++) = 0;
-	if (path_ptr)
-	  tp += 3;
-	else
-	  tp += 2;
+	CLASSTYPE_MARKED5 (*tp) = 0;
+	tp += increment;
+      }
+
+    /* Now, guarantee that we are following the leftmost
+       path in the chain.  */
+    if (rval > 0
+	&& (DECL_OFFSET (TYPE_NAME (type)) != 0
+	    || TREE_VIA_VIRTUAL (type)))
+      {
+	/* Reduce all types yet to be fully processed into
+	   the base type we are looking for, or NULL_TREE.  */
+	for (tp = search_stack->first; tp < search_tail; tp += increment)
+	  {
+	    tree *sub_tp, sub_path_ptr;
+	    int sub_rval;
+
+	    /* Don't chase down more right-most paths.  */
+	    if (TREE_VIA_VIRTUAL (*tp)
+		|| DECL_OFFSET (TYPE_NAME (*tp)) > DECL_OFFSET (TYPE_NAME (type)))
+	      {
+		*tp = NULL_TREE;
+		continue;
+	      }
+
+	    /* Don't hassle with duplicates.  */
+	    if (*tp == type)
+	      goto skip;
+
+	    for (sub_tp = search_stack->first; sub_tp < tp; sub_tp += increment)
+	      if (*tp == *sub_tp)
+		goto skip;
+
+	    /* Find this type's TYPE basetype, if it has one.  */
+	    sub_rval = get_base_distance (parent, *tp, 0, &sub_path_ptr);
+	    if (sub_rval == -1)
+	      *tp = NULL_TREE;
+	    else
+	      {
+		if (path_ptr && TREE_CHAIN (tp[3]))
+		  {
+		    tree last;
+		    tree next_to_last = sub_path_ptr;
+		    while (TREE_CHAIN (next_to_last)
+			   && TREE_CHAIN (TREE_CHAIN (next_to_last)))
+		      next_to_last = TREE_CHAIN (next_to_last);
+		    if (next_to_last == sub_path_ptr)
+		      {
+			sub_path_ptr = copy_node (sub_path_ptr);
+			last = sub_path_ptr;
+		      }
+		    else
+		      {
+			last = copy_node (TREE_CHAIN (next_to_last));
+			TREE_CHAIN (next_to_last) = last;
+		      }
+		    TREE_CHAIN (last) = TREE_CHAIN (tp[3]);
+		  }
+		*tp = TREE_VALUE (sub_path_ptr);
+		if (path_ptr)
+		  tp[3] = sub_path_ptr;
+	      }
+	  skip: {}
+	  }
+
+	/* For all the types which reduce to TYPE, choose
+	   the leftmost non-virtual one of them.  */
+	for (tp = search_stack->first; tp < search_tail; tp += increment)
+	  {
+	    if (*tp == NULL_TREE)
+	      continue;
+
+	    if (DECL_OFFSET (TYPE_NAME (*tp)) < DECL_OFFSET (TYPE_NAME (type)))
+	      {
+		rval = -2;
+		type = *tp;
+		if (path_ptr)
+		  basetypes = tp[3];
+	      }
+	  }
+	if (rval == -2)
+	  rval_private = 0;
       }
   }
   search_stack = pop_search_level (search_stack);
@@ -1803,10 +1885,10 @@ get_abstract_virtuals (type)
     }
   for (vbases = CLASSTYPE_VBASECLASSES (type); vbases; vbases = TREE_CHAIN (vbases))
     {
-      base = TREE_TYPE (vbases);
-      if (! TYPE_VIRTUAL_P (base))
+      if (! ASSOC_VIRTUALS (vbases));
 	continue;
 
+      base = TREE_TYPE (vbases);
       tmp = TREE_CHAIN (ASSOC_VIRTUALS (vbases));
       while (tmp)
 	{
@@ -2111,7 +2193,7 @@ dfs_init_vbase_pointers (type)
 }
 
 /* Sometimes this needs to clear both 3 and 4.  Other times,
-   just 4, but optimizer should make both equivalent
+   just 4, but optimizer should make both with equal efficiency
    (though it does not currently).  */
 static void
 dfs_clear_vbase_slots (type)
@@ -2933,23 +3015,32 @@ build_type_pathname (format, parent, type)
      tree parent, type;
 {
   extern struct obstack temporary_obstack;
-  char *base = obstack_base (&temporary_obstack);
-  char *name;
+  char *first, *base, *name;
   int i;
   tree id;
 
   parent = TYPE_MAIN_VARIANT (parent);
+
+  /* Remember where to cut the obstack to.  */
+  first = obstack_base (&temporary_obstack);
+
+  /* Put on TYPE+PARENT.  */
   obstack_grow (&temporary_obstack,
 		TYPE_NAME_STRING (type), TYPE_NAME_LENGTH (type));
   obstack_1grow (&temporary_obstack, JOINER);
   obstack_grow0 (&temporary_obstack,
 		 TYPE_NAME_STRING (parent), TYPE_NAME_LENGTH (parent));
   i = obstack_object_size (&temporary_obstack);
-  name = obstack_base (&temporary_obstack) + i;
+  base = obstack_base (&temporary_obstack);
+  obstack_finish (&temporary_obstack);
+
+  /* Put on FORMAT+TYPE+PARENT.  */
   obstack_blank (&temporary_obstack, strlen (format) + i + 1);
+  name = obstack_base (&temporary_obstack);
   sprintf (name, format, base);
   id = get_identifier (name);
-  obstack_free (&temporary_obstack, base);
+  obstack_free (&temporary_obstack, first);
+
   return id;
 }
 
