@@ -956,7 +956,7 @@ build_x_arrow (datum)
   if (last_rval == error_mark_node)
     return error_mark_node;
 
-  while (rval = build_opfncall (COMPONENT_REF, rval))
+  while (rval = build_opfncall (COMPONENT_REF, LOOKUP_NORMAL, rval))
     {
       if (rval == error_mark_node)
 	return error_mark_node;
@@ -983,7 +983,7 @@ build_x_arrow (datum)
 
   if (TREE_CODE (TREE_TYPE (last_rval)) == OFFSET_TYPE)
     {
-      if (TREE_CODE (last_rval) == MEMBER_REF
+      if (TREE_CODE (last_rval) == OFFSET_REF
 	  && TREE_STATIC (TREE_OPERAND (last_rval, 1)))
 	{
 	  last_rval = TREE_OPERAND (last_rval, 1);
@@ -1038,7 +1038,7 @@ build_m_component_ref (datum, component)
       return error_mark_node;
     }
 
-  return build (MEMBER_REF, TREE_TYPE (TREE_TYPE (component)), datum, component);
+  return build (OFFSET_REF, TREE_TYPE (TREE_TYPE (component)), datum, component);
 }
 
 /* Return a tree node for the expression TYPENAME '(' PARMS ')'.  */
@@ -1172,9 +1172,7 @@ build_functional_cast (exp, parms)
 
     if (rval != error_mark_node)
       {
-	rval = build (NEW_EXPR, type,
-		      TREE_OPERAND (rval, 0), TREE_OPERAND (rval, 1), 0);
-	TREE_ADDRESSABLE (rval) = 1;
+	rval = build_cplus_new (type, rval);
 	return rval;
       }
 
