@@ -905,6 +905,8 @@ dbxout_symbol (decl, local)
 	 and are output by dbxout_reg_parms and dbxout_parms.  */
       abort ();
 
+    case RESULT_DECL:
+      /* Named return value, treat like a VAR_DECL.  */
     case VAR_DECL:
       if (DECL_RTL (decl) == 0)
 	return;
@@ -1177,15 +1179,12 @@ dbxout_parms (parms)
 	  if (DECL_NAME (parms))
 	    {
 	      current_sym_nchars = 2 + IDENTIFIER_LENGTH (DECL_NAME (parms));
-
-	      FORCE_TEXT;
 	      fprintf (asmfile, ".stabs \"%s:P",
 		       IDENTIFIER_POINTER (DECL_NAME (parms)));
 	    }
 	  else
 	    {
 	      current_sym_nchars = 8;
-	      FORCE_TEXT;
 	      fprintf (asmfile, ".stabs \"(anon):P");
 	    }
 
@@ -1205,15 +1204,12 @@ dbxout_parms (parms)
 	  if (DECL_NAME (parms))
 	    {
 	      current_sym_nchars = 2 + IDENTIFIER_LENGTH (DECL_NAME (parms));
-
-	      FORCE_TEXT;
 	      fprintf (asmfile, ".stabs \"%s:p",
 		       IDENTIFIER_POINTER (DECL_NAME (parms)));
 	    }
 	  else
 	    {
 	      current_sym_nchars = 8;
-	      FORCE_TEXT;
 	      fprintf (asmfile, ".stabs \"(anon):p");
 	    }
 
@@ -1264,14 +1260,12 @@ dbxout_reg_parms (parms)
 	  if (DECL_NAME (parms))
 	    {
 	      current_sym_nchars = 2 + IDENTIFIER_LENGTH (DECL_NAME (parms));
-	      FORCE_TEXT;
 	      fprintf (asmfile, ".stabs \"%s:r",
 		       IDENTIFIER_POINTER (DECL_NAME (parms)));
 	    }
 	  else
 	    {
 	      current_sym_nchars = 8;
-	      FORCE_TEXT;
 	      fprintf (asmfile, ".stabs \"(anon):r");
 	    }
 	  dbxout_type (TREE_TYPE (parms), 0);
@@ -1300,14 +1294,12 @@ dbxout_reg_parms (parms)
 	      if (DECL_NAME (parms))
 		{
 		  current_sym_nchars = 2 + IDENTIFIER_LENGTH (DECL_NAME (parms));
-		  FORCE_TEXT;
 		  fprintf (asmfile, ".stabs \"%s:",
 			   IDENTIFIER_POINTER (DECL_NAME (parms)));
 		}
 	      else
 		{
 		  current_sym_nchars = 8;
-		  FORCE_TEXT;
 		  fprintf (asmfile, ".stabs \"(anon):");
 		}
 	      dbxout_type (TREE_TYPE (parms), 0);
@@ -1486,8 +1478,12 @@ void
 dbxout_function (decl)
      tree decl;
 {
+  extern tree value_identifier;
+
   dbxout_symbol (decl, 0);
   dbxout_parms (DECL_ARGUMENTS (decl));
+  if (DECL_NAME (DECL_RESULT (decl)) != value_identifier)
+    dbxout_symbol (DECL_RESULT (decl), 1);
   dbxout_block (DECL_INITIAL (decl), 0, DECL_ARGUMENTS (decl));
   
   /* If we made any temporary types in this fn that weren't
