@@ -1,6 +1,5 @@
-/* Language-indepednent node constructors for parse phase of GNU compiler.
+/* Language-independent node constructors for parse phase of GNU compiler.
    Copyright (C) 1987, 1988 Free Software Foundation, Inc.
-   Hacked by Michael Tiemann (tiemann@mcc.com)
 
 This file is part of GNU CC.
 
@@ -36,9 +35,8 @@ the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */
 #include "config.h"
 #include <stdio.h>
 #include "tree.h"
-#include "cplus-tree.h"
 #include "obstack.h"
-#include "varargs.h"
+#include "gvarargs.h"
 #include "flags.h"
 
 #define obstack_chunk_alloc xmalloc
@@ -69,6 +67,12 @@ struct obstack temporary_obstack;
 
 struct obstack momentary_obstack;
 
+/* The tree nodes of a declarator are allocated
+   in this obstack, and all are freed when the declarator
+   has been parsed.  */
+
+static struct obstack temp_decl_obstack;
+
 /* This points at either permanent_obstack or maybepermanent_obstack.  */
 
 struct obstack *saveable_obstack;
@@ -93,6 +97,7 @@ struct obstack *expression_obstack;
 char *maybepermanent_firstobj;
 char *temporary_firstobj;
 char *momentary_firstobj;
+char *temp_decl_firstobj;
 
 /* Nonzero means all ..._TYPE nodes should be allocated permanently.  */
 
@@ -118,7 +123,7 @@ struct momentary_level *momentary_stack;
 
 #define DEFTREECODE(SYM, NAME, TYPE, LENGTH) TYPE,
 
-char *tree_code_type[] = {
+char *standard_tree_code_type[] = {
 #include "tree.def"
 };
 #undef DEFTREECODE
@@ -129,27 +134,61 @@ char *tree_code_type[] = {
 
 #define DEFTREECODE(SYM, NAME, TYPE, LENGTH) LENGTH,
 
-int tree_code_length[] = {
+int standard_tree_code_length[] = {
 #include "tree.def"
 };
 #undef DEFTREECODE
 
+/* Names of tree components.
+   Used for printing out the tree and error messages.  */
+#define DEFTREECODE(SYM, NAME, TYPE, LEN) NAME,
+
+char *standard_tree_code_name[] = {
+#include "tree.def"
+};
+#undef DEFTREECODE
+
+/* Table indexed by tree code giving a string containing a character
+   classifying the tree code.  Possibilities are
+   t, d, s, c, r and e.  See tree.def for details.  */
+
+char **tree_code_type;
+
+/* Table indexed by tree code giving number of expression
+   operands beyond the fixed part of the node structure.
+   Not used for types or decls.  */
+
+int *tree_code_length;
+
+/* Table indexed by tree code giving name of tree code, as a string.  */
+
+char **tree_code_name;
+
 /* Counter for assigning unique ids to all tree nodes.  */
 
 int tree_node_counter = 0;
+
+/* Statistics-gathering stuff.  */
 typedef enum
 {
-  d_kind, t_kind, s_kind, r_kind, e_kind, c_kind, x_kind,
-  lang_decl, lang_type, all_kinds
+  d_kind, t_kind, s_kind, r_kind, e_kind, c_kind,
+  id_kind, op_id_kind, perm_list_kind, temp_list_kind,
+  x_kind, lang_decl, lang_type, all_kinds
 } tree_node_kind;
 int tree_node_kinds[(int)all_kinds];
+int tree_node_sizes[(int)all_kinds];
+int id_string_size = 0;
 char *tree_node_kind_names[] = { "decls", "types", "stmts", "refs", "exprs", "constants",
+				 "identifiers", "op_identifiers", "perm_tree_lists", "temp_tree_lists",
 				 "random kinds", "lang_decl kinds", "lang_type kinds" };
 
 /* Hash table for uniquizing IDENTIFIER_NODEs by name.  */
 
 #define MAX_HASH_TABLE 1009
 static tree hash_table[MAX_HASH_TABLE];	/* id hash buckets */
+
+/* 0 while creating built-in identifiers.  */
+static int do_identifier_warnings;
 
 /* Init data for node creation, at the beginning of compilation.  */
 
@@ -165,14 +204,76 @@ init_tree ()
   obstack_init (&maybepermanent_obstack);
   maybepermanent_firstobj
     = (char *) obstack_alloc (&maybepermanent_obstack, 0);
+  obstack_init (&temp_decl_obstack);
+  temp_decl_firstobj = (char *) obstack_alloc (&temp_decl_obstack, 0);
 
   current_obstack = &permanent_obstack;
   expression_obstack = &permanent_obstack;
   rtl_obstack = saveable_obstack = &permanent_obstack;
   tree_node_counter = 1;
-  bzero (hash_table, sizeof hash_table);
+  /* bzero (hash_table, sizeof hash_table); */
+
+  tree_code_type = (char **) xmalloc (sizeof (standard_tree_code_type));
+  tree_code_length = (int *) xmalloc (sizeof (standard_tree_code_length));
+  tree_code_name = (char **) xmalloc (sizeof (standard_tree_code_name));
+  bcopy (standard_tree_code_type, tree_code_type,
+	 sizeof (standard_tree_code_type));
+  bcopy (standard_tree_code_length, tree_code_length,
+	 sizeof (standard_tree_code_length));
+  bcopy (standard_tree_code_name, tree_code_name,
+	 sizeof (standard_tree_code_name));
+}
+#if 0
+/* Save all variables describing the current status into the structure *P.
+   This is used before starting a nested function.  */
+
+void
+save_tree_status (p)
+     struct function *p;
+{
+  p->all_types_permanent = all_types_permanent;
+  p->momentary_stack = momentary_stack;
+  p->maybepermanent_firstobj = maybepermanent_firstobj;
+  p->temporary_firstobj = temporary_firstobj;
+  p->momentary_firstobj = momentary_firstobj;
+  p->current_obstack = current_obstack;
+  p->expression_obstack = expression_obstack;
+  p->saveable_obstack = saveable_obstack;
+  p->rtl_obstack = rtl_obstack;
+
+  current_obstack = &permanent_obstack;
+  expression_obstack = &permanent_obstack;
+  rtl_obstack = saveable_obstack = &permanent_obstack;
+
+  maybepermanent_firstobj = (char *) obstack_finish (&maybepermanent_obstack);
+  temporary_firstobj = (char *) obstack_finish (&temporary_obstack);
+  momentary_firstobj = (char *) obstack_finish (&momentary_obstack);
 }
 
+/* Restore all variables describing the current status from the structure *P.
+   This is used after a nested function.  */
+
+void
+restore_tree_status (p)
+     struct function *p;
+{
+  all_types_permanent = p->all_types_permanent;
+  momentary_stack = p->momentary_stack;
+
+  obstack_free (&maybepermanent_obstack, maybepermanent_firstobj);
+  obstack_free (&temporary_obstack, temporary_firstobj);
+  obstack_free (&momentary_obstack, momentary_firstobj);
+
+  maybepermanent_firstobj = p->maybepermanent_firstobj;
+  temporary_firstobj = p->temporary_firstobj;
+  momentary_firstobj = p->momentary_firstobj;
+  current_obstack = p->current_obstack;
+  expression_obstack = p->expression_obstack;
+  saveable_obstack = p->saveable_obstack;
+  rtl_obstack = p->rtl_obstack;
+}
+#endif
+
 /* Start allocating on the temporary (per function) obstack.
    This is done in start_function before parsing the function body,
    and before each initialization at top level, and to go back
@@ -230,6 +331,7 @@ permanent_allocation ()
   obstack_free (&temporary_obstack, temporary_firstobj);
   obstack_free (&momentary_obstack, momentary_firstobj);
   obstack_free (&maybepermanent_obstack, maybepermanent_firstobj);
+  obstack_free (&temp_decl_obstack, temp_decl_firstobj);
 
   current_obstack = &permanent_obstack;
   expression_obstack = &permanent_obstack;
@@ -244,6 +346,17 @@ preserve_data ()
   maybepermanent_firstobj
     = (char *) obstack_alloc (&maybepermanent_obstack, 0);
 }
+
+void
+preserve_initializer ()
+{
+  temporary_firstobj
+    = (char *) obstack_alloc (&temporary_obstack, 0);
+  momentary_firstobj
+    = (char *) obstack_alloc (&momentary_obstack, 0);
+  maybepermanent_firstobj
+    = (char *) obstack_alloc (&maybepermanent_obstack, 0);
+}
 
 /* Allocate SIZE bytes in the current obstack
    and return a pointer to them.
@@ -276,6 +389,16 @@ permalloc (size)
 {
   return (char *) obstack_alloc (&permanent_obstack, size);
 }
+
+/* Allocate SIZE bytes in the saveable obstack
+   and return a pointer to them.  */
+
+char *
+savealloc (size)
+     int size;
+{
+  return (char *) obstack_alloc (saveable_obstack, size);
+}
 
 /* Start a level of momentary allocation.
    In C, each compound statement has its own level
@@ -359,72 +482,94 @@ make_node (code)
   register int length;
   register struct obstack *obstack = current_obstack;
   register int i;
+  register tree_node_kind kind;
 
   switch (type)
     {
     case 'd':  /* A decl node */
-      tree_node_kinds[(int)d_kind]++;
-      length = sizeof (struct tree_decl);
+      length = sizeof (struct tree_decl)
+	+ tree_code_length[(int) code] * sizeof (char *);
+      kind = d_kind;
       /* All decls in an inline function need to be saved.  */
       if (obstack != &permanent_obstack)
 	obstack = saveable_obstack;
+      /* PARM_DECLs always go on the saveable_obstack, not permanent
+	 even though we may make them before the function turns
+	 on temporary allocation.  */
+      else if (code == PARM_DECL)
+	obstack = &maybepermanent_obstack;
       break;
 
     case 't':  /* a type node */
-      tree_node_kinds[(int)t_kind]++;
       length = sizeof (struct tree_type);
+      kind = t_kind;
       /* All data types are put where we can preserve them if nec.  */
       if (obstack != &permanent_obstack)
 	obstack = all_types_permanent ? &permanent_obstack : saveable_obstack;
       break;
 
     case 's':  /* a stmt node */
-      tree_node_kinds[(int)s_kind]++;
       length = sizeof (struct tree_common)
 	+ 2 * sizeof (int)
 	  + tree_code_length[(int) code] * sizeof (char *);
+      kind = s_kind;
       /* All stmts are put where we can preserve them if nec.  */
       if (obstack != &permanent_obstack)
 	obstack = saveable_obstack;
       break;
 
     case 'r':  /* a reference */
-      tree_node_kinds[(int)r_kind]++;
-      tree_node_kinds[(int)e_kind]++;
       obstack = expression_obstack;
       length = sizeof (struct tree_exp)
 	+ (tree_code_length[(int) code] - 1) * sizeof (char *);
+      kind = r_kind;
       break;
 
     case 'e':  /* an expression */
-      tree_node_kinds[(int)e_kind]++;
       obstack = expression_obstack;
       length = sizeof (struct tree_exp)
 	+ (tree_code_length[(int) code] - 1) * sizeof (char *);
+      kind = e_kind;
       break;
 
     case 'c':  /* a constant */
-      tree_node_kinds[(int)c_kind]++;
       obstack = expression_obstack;
       /* We can't use tree_code_length for this, since the number of words
 	 is machine-dependent due to varying alignment of `double'.  */
       if (code == REAL_CST)
 	{
 	  length = sizeof (struct tree_real_cst);
+	  kind = c_kind;
 	  break;
 	}
+      length = sizeof (struct tree_common)
+	+ tree_code_length[(int) code] * sizeof (char *);
+      kind = c_kind;
+      break;
 
     case 'x':  /* something random, like an identifier.  */
-      tree_node_kinds[(int)x_kind]++;
       length = sizeof (struct tree_common)
 	+ tree_code_length[(int) code] * sizeof (char *);
       /* Identifier nodes are always permanent since they are
 	 unique in a compiler run.  */
-      if (code == IDENTIFIER_NODE) obstack = &permanent_obstack;
+      if (code == IDENTIFIER_NODE)
+	{
+	  kind = id_kind;
+	  obstack = &permanent_obstack;
+	}
+      else if (code == OP_IDENTIFIER)
+	kind = op_id_kind;
+      else
+	kind = x_kind;
     }
 
   t = (tree) obstack_alloc (obstack, length);
 
+#ifdef GATHER_STATISTICS
+  tree_node_kinds[(int)kind]++;
+  tree_node_sizes[(int)kind] += length;
+#endif
+
   TREE_UID (t) = tree_node_counter++;
   TREE_TYPE (t) = 0;
   TREE_CHAIN (t) = 0;
@@ -444,8 +589,14 @@ make_node (code)
       DECL_ALIGN (t) = 1;
       DECL_SIZE_UNIT (t) = 1;
       DECL_VOFFSET_UNIT (t) = 1;
-      DECL_SOURCE_LINE (t) = lineno;
-      DECL_SOURCE_FILE (t) = input_filename;
+      if (code == PARM_DECL)
+	DECL_CONTEXT (t) = current_function_decl;
+      else
+	{
+	  DECL_SOURCE_LINE (t) = lineno;
+	  DECL_SOURCE_FILE (t)
+	    = (input_filename) ? input_filename : "<built-in>";
+	}
     }
 
   if (type == 't')
@@ -478,7 +629,8 @@ copy_node (node)
   switch (*tree_code_type[(int) code])
     {
     case 'd':  /* A decl node */
-      length = sizeof (struct tree_decl);
+      length = sizeof (struct tree_decl)
+	+ tree_code_length[(int) code] * sizeof (char *);
       break;
 
     case 't':  /* a type node */
@@ -509,11 +661,13 @@ copy_node (node)
     case 'x':  /* something random, like an identifier.  */
       length = sizeof (struct tree_common)
 	+ tree_code_length[(int) code] * sizeof (char *);
+      if (code == TREE_VEC)
+	length += (TREE_VEC_LENGTH (node) - 1) * sizeof (char *);
     }
 
   t = (tree) obstack_alloc (current_obstack, length);
 
-  for (i = (length / sizeof (int)) - 1;
+  for (i = ((length + sizeof (int) - 1) / sizeof (int)) - 1;
        i >= 0;
        i--)
     ((int *) t)[i] = ((int *) node)[i];
@@ -526,8 +680,8 @@ copy_node (node)
   return t;
 }
 
-/* Return a new list which contains copies of the nodes found
-   in LIST via the TREE_CHAIN field.  */
+/* Return a copy of a chain of nodes, chained through the TREE_CHAIN field.
+   For example, this can copy a list made of TREE_LIST nodes.  */
 
 tree
 copy_list (list)
@@ -574,8 +728,8 @@ get_identifier (text)
     hash_len = id_clash_len;
 
   /* Compute hash code */
-  hi = hash_len;
-  for (i = 0; i < hash_len; i++)
+  hi = 17 * (unsigned)(text[0]);
+  for (i = 1; i < hash_len; i += 2)
     hi = ((hi * 613) + (unsigned)(text[i]));
 
   hi &= (1 << HASHBITS) - 1;
@@ -584,11 +738,12 @@ get_identifier (text)
   /* Search table for identifier */
   for (idp = hash_table[hi]; idp; idp = TREE_CHAIN (idp))
     if (IDENTIFIER_LENGTH (idp) == len
-	&& !strcmp (IDENTIFIER_POINTER (idp), text))
+	&& IDENTIFIER_POINTER (idp)[0] == text[0]
+	&& !bcmp (IDENTIFIER_POINTER (idp), text, len))
       return idp;		/* <-- return if found */
-  
+
   /* Not found; optionally warn about a similar identifier */
-  if (warn_id_clash && len > id_clash_len)
+  if (warn_id_clash && do_identifier_warnings && len > id_clash_len)
     for (idp = hash_table[hi]; idp; idp = TREE_CHAIN (idp))
       if (!strncmp (IDENTIFIER_POINTER (idp), text, id_clash_len))
 	{
@@ -597,9 +752,13 @@ get_identifier (text)
 	  break;
 	}
 
+  if (tree_code_length[(int) IDENTIFIER_NODE] < 0)
+    abort ();			/* set_identifier_size hasn't been called.  */
+
   /* Not found, create one, add to chain */
   idp = make_node (IDENTIFIER_NODE);
   IDENTIFIER_LENGTH (idp) = len;
+  id_string_size += len;
 
   IDENTIFIER_POINTER (idp) = obstack_copy0 (&permanent_obstack, text, len);
 
@@ -607,6 +766,28 @@ get_identifier (text)
   hash_table[hi] = idp;
   return idp;			/* <-- return if created */
 }
+
+/* Enable warnings on similar identifiers (if requested).
+   Done after the built-in identifiers are created.  */
+
+void
+start_identifier_warnings ()
+{
+  do_identifier_warnings = 1;
+}
+
+/* Record the size of an identifier node for the language in use.
+   SIZE is the total size in bytes.
+   This is called by the language-specific files.  This must be
+   called before allocating any identifiers.  */
+
+void
+set_identifier_size (size)
+     int size;
+{
+  tree_code_length[(int) IDENTIFIER_NODE]
+    = (size - sizeof (struct tree_common)) / sizeof (tree);
+}
 
 /* Return a newly constructed INTEGER_CST node whose constant value
    is specified by the two ints LOW and HI.
@@ -616,27 +797,7 @@ tree
 build_int_2 (low, hi)
      int low, hi;
 {
-#if 0
   register tree t = make_node (INTEGER_CST);
-#else
-  register tree t;
-  register struct obstack *obstack = current_obstack;
-  register int i, length;
-
-  tree_node_kinds[(int)c_kind]++;
-  length = sizeof (struct tree_common) + 2 * sizeof (char *);
-
-  t = (tree) obstack_alloc (obstack, length);
-  TREE_UID (t) = tree_node_counter++;
-  TREE_TYPE (t) = 0;
-  TREE_CHAIN (t) = 0;
-  ((int *) t)[3] = 0;
-
-  TREE_SET_CODE (t, INTEGER_CST);
-  if (obstack == &permanent_obstack)
-    TREE_PERMANENT (t) = 1;
-#endif
-  TREE_LITERAL (t) = 1;
   TREE_INT_CST_LOW (t) = low;
   TREE_INT_CST_HIGH (t) = hi;
   TREE_TYPE (t) = integer_type_node;
@@ -668,20 +829,12 @@ build_real (type, d)
    and whose value is the integer value of the INTEGER_CST node I.  */
 
 #if !defined (REAL_IS_NOT_DOUBLE) || defined (REAL_ARITHMETIC)
-/* This function can't be implemented if we can't do arithmetic
-   on the float representation.  */
 
-tree
-build_real_from_int_cst (type, i)
-     tree type;
+REAL_VALUE_TYPE
+real_value_from_int_cst (i)
      tree i;
 {
-  tree v;
-  double d;
-
-  v = make_node (REAL_CST);
-  TREE_TYPE (v) = type;
-
+  REAL_VALUE_TYPE d;
 #ifdef REAL_ARITHMETIC
   REAL_VALUE_FROM_INT (d, TREE_INT_CST_LOW (i), TREE_INT_CST_HIGH (i));
 #else /* not REAL_ARITHMETIC */
@@ -701,7 +854,24 @@ build_real_from_int_cst (type, i)
       d += (double) (unsigned) TREE_INT_CST_LOW (i);
     }
 #endif /* not REAL_ARITHMETIC */
+  return d;
+}
 
+/* This function can't be implemented if we can't do arithmetic
+   on the float representation.  */
+
+tree
+build_real_from_int_cst (type, i)
+     tree type;
+     tree i;
+{
+  tree v;
+  REAL_VALUE_TYPE d;
+
+  v = make_node (REAL_CST);
+  TREE_TYPE (v) = type;
+
+  d = real_value_from_int_cst (i);
   /* Check for valid float value for this type on this target machine;
      if not, can print error message and store a valid value in D.  */
 #ifdef CHECK_FLOAT_VALUE
@@ -743,6 +913,38 @@ build_complex (real, imag)
   TREE_IMAGPART (t) = imag;
   return t;
 }
+
+/* Build a newly constructed TREE_VEC node of length LEN.  */
+tree
+make_tree_vec (len)
+     int len;
+{
+  register tree t;
+  register int length = (len-1) * sizeof (tree) + sizeof (struct tree_vec);
+  register struct obstack *obstack = current_obstack;
+  register int i;
+
+#ifdef GATHER_STATISTICS
+  tree_node_kinds[(int)x_kind]++;
+  tree_node_sizes[(int)x_kind] += length;
+#endif
+
+  t = (tree) obstack_alloc (obstack, length);
+
+  TREE_UID (t) = tree_node_counter++;
+  TREE_TYPE (t) = 0;
+  TREE_CHAIN (t) = 0;
+  for (i = (length / sizeof (int)) - 1;
+       i >= sizeof (struct tree_common) / sizeof (int) - 1;
+       i--)
+    ((int *) t)[i] = 0;
+  TREE_SET_CODE (t, TREE_VEC);
+  TREE_VEC_LENGTH (t) = len;
+  if (obstack == &permanent_obstack)
+    TREE_PERMANENT (t) = 1;
+
+  return t;
+}
 
 /* Return 1 if EXPR is the integer constant zero.  */
 
@@ -791,6 +993,36 @@ integer_all_onesp (expr)
     return TREE_INT_CST_LOW (expr) == (1 << prec) - 1;
 }
 
+/* Return list element whose TREE_VALUE is  ELEM.
+   Return 0 if ELEM is not in LIST.  */
+tree
+value_member (elem, list)
+     tree elem, list;
+{
+  while (list)
+    {
+      if (elem == TREE_VALUE (list))
+	return list;
+      list = TREE_CHAIN (list);
+    }
+  return NULL_TREE;
+}
+
+/* Return list element whose TREE_PURPOSE is  ELEM.
+   Return 0 if ELEM is not in LIST.  */
+tree
+purpose_member (elem, list)
+     tree elem, list;
+{
+  while (list)
+    {
+      if (elem == TREE_PURPOSE (list))
+	return list;
+      list = TREE_CHAIN (list);
+    }
+  return NULL_TREE;
+}
+
 /* Return the length of a chain of nodes chained through TREE_CHAIN.
    We expect a null pointer to mark the end of the chain.
    This is the Lisp primitive `length'.  */
@@ -830,36 +1062,30 @@ chainon (op1, op2)
 
 /* Return a newly created TREE_LIST node whose
    purpose and value fields are PARM and VALUE.  */
+
 tree
 build_tree_list (parm, value)
      tree parm, value;
 {
-#if 0
   register tree t = make_node (TREE_LIST);
-#else
-  register tree t;
-  register struct obstack *obstack = current_obstack;
-  register int i, length;
-
-  tree_node_kinds[(int)x_kind]++;
-  length = sizeof (struct tree_common) + 2 * sizeof (char *);
-
-  t = (tree) obstack_alloc (obstack, length);
-  TREE_UID (t) = tree_node_counter++;
-  TREE_TYPE (t) = 0;
-  TREE_CHAIN (t) = 0;
-  ((int *) t)[3] = 0;
-
-  TREE_SET_CODE (t, TREE_LIST);
-  if (obstack == &permanent_obstack)
-    TREE_PERMANENT (t) = 1;
-#endif
-
   TREE_PURPOSE (t) = parm;
   TREE_VALUE (t) = value;
   return t;
 }
 
+/* Similar, but build on the temp_decl_obstack.  */
+tree
+build_decl_list (parm, value)
+     tree parm, value;
+{
+  register tree node;
+  register struct obstack *ambient_obstack = current_obstack;
+  current_obstack = &temp_decl_obstack;
+  node = build_tree_list (parm, value);
+  current_obstack = ambient_obstack;
+  return node;
+}
+
 /* Return a newly created TREE_LIST node whose
    purpose and value fields are PARM and VALUE
    and whose TREE_CHAIN is CHAIN.  */
@@ -868,29 +1094,24 @@ tree
 tree_cons (purpose, value, chain)
      tree purpose, value, chain;
 {
-#if 0
-  register tree t = make_node (TREE_LIST);
-#else
-  register tree t;
-  register struct obstack *obstack = current_obstack;
-  register int i, length;
+  register tree node = make_node (TREE_LIST);
+  TREE_CHAIN (node) = chain;
+  TREE_PURPOSE (node) = purpose;
+  TREE_VALUE (node) = value;
+  return node;
+}
 
-  tree_node_kinds[(int)x_kind]++;
-  length = sizeof (struct tree_common) + 2 * sizeof (char *);
-
-  t = (tree) obstack_alloc (obstack, length);
-  TREE_UID (t) = tree_node_counter++;
-  TREE_TYPE (t) = 0;
-  ((int *) t)[3] = 0;
-
-  TREE_SET_CODE (t, TREE_LIST);
-  if (obstack == &permanent_obstack)
-    TREE_PERMANENT (t) = 1;
-#endif
-  TREE_CHAIN (t) = chain;
-  TREE_PURPOSE (t) = purpose;
-  TREE_VALUE (t) = value;
-  return t;
+/* Similar, but build on the temp_decl_obstack.  */
+tree
+decl_tree_cons (purpose, value, chain)
+     tree purpose, value, chain;
+{
+  register tree node;
+  register struct obstack *ambient_obstack = current_obstack;
+  current_obstack = &temp_decl_obstack;
+  node = tree_cons (purpose, value, chain);
+  current_obstack = ambient_obstack;
+  return node;
 }
 
 /* Same as `tree_cons' but make a permanent object.  */
@@ -904,6 +1125,12 @@ perm_tree_cons (purpose, value, chain)
   current_obstack = &permanent_obstack;
 
   node = make_node (TREE_LIST);
+#ifdef GATHER_STATISTICS
+  tree_node_kinds[x_kind]--;
+  tree_node_sizes[(int)x_kind] -= sizeof(struct tree_list);
+  tree_node_kinds[perm_list_kind]++;
+  tree_node_sizes[(int)perm_list_kind] += sizeof(struct tree_list);
+#endif
   TREE_CHAIN (node) = chain;
   TREE_PURPOSE (node) = purpose;
   TREE_VALUE (node) = value;
@@ -923,6 +1150,12 @@ temp_tree_cons (purpose, value, chain)
   current_obstack = &temporary_obstack;
 
   node = make_node (TREE_LIST);
+#ifdef GATHER_STATISTICS
+  tree_node_kinds[x_kind]--;
+  tree_node_kinds[temp_list_kind]++;
+  tree_node_sizes[(int)x_kind] -= sizeof(struct tree_list);
+  tree_node_sizes[(int)temp_list_kind] += sizeof(struct tree_list);
+#endif
   TREE_CHAIN (node) = chain;
   TREE_PURPOSE (node) = purpose;
   TREE_VALUE (node) = value;
@@ -979,6 +1212,31 @@ nreverse (t)
     }
   return prev;
 }
+
+/* Given a chain CHAIN of tree nodes,
+   construct and return a list of those nodes.  */
+
+tree
+listify (chain)
+     tree chain;
+{
+  tree result = NULL_TREE;
+  tree in_tail = chain;
+  tree out_tail = NULL_TREE;
+
+  while (in_tail)
+    {
+      tree next = tree_cons (NULL_TREE, in_tail, NULL_TREE);
+      if (out_tail)
+	TREE_CHAIN (out_tail) = next;
+      else
+	result = next;
+      out_tail = next;
+      in_tail = TREE_CHAIN (in_tail);
+    }
+
+  return result;
+}
 
 /* Return the size nominally occupied by an object of type TYPE
    when it resides in memory.  The value is measured in units of bytes,
@@ -1068,70 +1326,6 @@ staticp (arg)
 
   return 0;
 }
-
-/* Return nonzero if REF is an lvalue valid for this language.
-   Lvalues can be assigned, unless they have TREE_READONLY.
-   Lvalues can have their address taken, unless they have TREE_REGDECL.  */
-
-int
-lvalue_p (ref)
-     tree ref;
-{
-  register enum tree_code code = TREE_CODE (ref);
-
-  if (language_lvalue_valid (ref))
-    switch (code)
-      {
-      case COMPONENT_REF:
-	return lvalue_p (TREE_OPERAND (ref, 0));
-
-      case STRING_CST:
-	return 1;
-
-      case INDIRECT_REF:
-      case ARRAY_REF:
-      case VAR_DECL:
-      case PARM_DECL:
-      case RESULT_DECL:
-      case ERROR_MARK:
-	if (TREE_CODE (TREE_TYPE (ref)) != FUNCTION_TYPE
-	    && TREE_CODE (TREE_TYPE (ref)) != METHOD_TYPE)
-	  return 1;
-	break;
-
-      case NEW_EXPR:
-	return 1;
-
-      case CALL_EXPR:
-	if (TREE_CODE (TREE_TYPE (ref)) == REFERENCE_TYPE)
-	  return 1;
-	break;
-
-	/* A currently unresolved scope ref.  */
-      case SCOPE_REF:
-	abort ();
-      case MEMBER_REF:
-	if (TREE_CODE (TREE_OPERAND (ref, 1)) == VAR_DECL
-	    || TREE_CODE (TREE_OPERAND (ref, 1)) == FUNCTION_DECL)
-	  return 1;
-	break;
-      }
-  return 0;
-}
-
-/* Return nonzero if REF is an lvalue valid for this language;
-   otherwise, print an error message and return zero.  */
-
-int
-lvalue_or_else (ref, string)
-     tree ref;
-     char *string;
-{
-  int win = lvalue_p (ref);
-  if (! win)
-    error ("invalid lvalue in %s", string);
-  return win;
-}
 
 /* This should be applied to any node which may be used in more than one place,
    but must be evaluated only once.  Normally, the code generator would
@@ -1267,6 +1461,57 @@ build (va_alist)
   return t;
 }
 
+/* Same as above, but only builds for unary operators.
+   Saves lions share of calls to `build'; cuts down use
+   of varargs, which is expensive for RISC machines.  */
+tree
+build1 (code, type, node)
+     enum tree_code code;
+     tree type;
+     tree node;
+{
+  register struct obstack *obstack = current_obstack;
+  register int i, length;
+  register tree_node_kind kind;
+  register tree t;
+
+  if (*tree_code_type[(int) code] == 'r')
+    kind = r_kind;
+  else if (*tree_code_type[(int) code] == 'e')
+    kind = e_kind;
+  else
+    abort ();
+
+  obstack = expression_obstack;
+  length = sizeof (struct tree_exp);
+
+  t = (tree) obstack_alloc (obstack, length);
+
+#ifdef GATHER_STATISTICS
+  tree_node_kinds[(int)kind]++;
+  tree_node_sizes[(int)kind] += length;
+#endif
+
+  TREE_UID (t) = tree_node_counter++;
+  TREE_TYPE (t) = type;
+  TREE_CHAIN (t) = 0;
+
+  for (i = (length / sizeof (int)) - 2;
+       i >= sizeof (struct tree_common) / sizeof (int) - 1;
+       i--)
+    ((int *) t)[i] = 0;
+  TREE_SET_CODE (t, code);
+
+  if (obstack == &permanent_obstack)
+    TREE_PERMANENT (t) = 1;
+
+  TREE_OPERAND (t, 0) = node;
+  if (node && TREE_VOLATILE (node))
+    TREE_VOLATILE (t) = 1;
+
+  return t;
+}
+
 /* Similar except don't specify the TREE_TYPE
    and leave the TREE_VOLATILE as 0.
    It is permissible for arguments to be null,
@@ -1295,6 +1540,39 @@ build_nt (va_alist)
   return t;
 }
 
+/* Similar to `build_nt', except we build
+   on the temp_decl_obstack, regardless.  */
+
+tree
+build_parse_node (va_alist)
+     va_dcl
+{
+  register struct obstack *ambient_obstack = expression_obstack;
+  register va_list p;
+  register enum tree_code code;
+  register tree t;
+  register int length;
+  register int i;
+
+  expression_obstack = &temp_decl_obstack;
+
+  va_start (p);
+
+  code = va_arg (p, enum tree_code);
+  t = make_node (code);
+  length = tree_code_length[(int) code];
+
+  for (i = 0; i < length; i++)
+    TREE_OPERAND (t, i) = va_arg (p, tree);
+
+  va_end (p);
+  expression_obstack = ambient_obstack;
+  return t;
+}
+
+#if 0
+/* Commented out because this wants to be done very
+   differently.  See cplus-lex.c.  */
 tree
 build_op_identifier (op1, op2)
      tree op1, op2;
@@ -1304,6 +1582,7 @@ build_op_identifier (op1, op2)
   TREE_VALUE (t) = op2;
   return t;
 }
+#endif
 
 /* Create a DECL_... node of code CODE, name NAME and data type TYPE.
    We do NOT enter this node in any sort of symbol table.
@@ -1326,9 +1605,21 @@ build_decl (code, name, type)
    as the type can suppress useless errors in the use of this variable.  */
 
   DECL_NAME (t) = name;
+  if (name)
+    {
+#if 0
+      DECL_PRINT_NAME (t) = IDENTIFIER_POINTER (name);
+#endif
+      if (code != PARM_DECL)
+	DECL_ASSEMBLER_NAME (t) = IDENTIFIER_POINTER (name);
+    }
   TREE_TYPE (t) = type;
+
+  /* A freshly built node has these properties anyway.  */
+#if 0
   DECL_ARGUMENTS (t) = NULL_TREE;
   DECL_INITIAL (t) = NULL_TREE;
+#endif /* 0 */
 
   if (code == VAR_DECL || code == PARM_DECL || code == RESULT_DECL)
     layout_decl (t, 0);
@@ -1338,7 +1629,6 @@ build_decl (code, name, type)
   return t;
 }
 
-#if 0
 /* Low-level constructors for statements.
    These constructors all expect source file name and line number
    as arguments, as well as enough arguments to fill in the data
@@ -1441,19 +1731,6 @@ build_case (filename, line, object, cases)
   return t;
 }
 
-tree
-build_loop (filename, line, body)
-     char *filename;
-     int line;
-     tree body;
-{
-  register tree t = make_node (LOOP_STMT);
-  STMT_SOURCE_FILE (t) = filename;
-  STMT_SOURCE_LINE (t) = line;
-  STMT_BODY (t) = body;
-  return t;
-}
-
 tree
 build_compound (filename, line, body)
      char *filename;
@@ -1467,23 +1744,36 @@ build_compound (filename, line, body)
   return t;
 }
 
-#endif /* 0 */
-
+tree
+build_loop (filename, line, vars, cond, body)
+     char *filename;
+     int line;
+     tree vars, cond, body;
+{
+  register tree t = make_node (LOOP_STMT);
+  STMT_SOURCE_FILE (t) = filename;
+  STMT_SOURCE_LINE (t) = line;
+  STMT_LOOP_VARS (t) = vars;
+  STMT_LOOP_COND (t) = cond;
+  STMT_LOOP_BODY (t) = body;
+  return t;
+}
+
 /* LET_STMT nodes are used to represent the structure of binding contours
    and declarations, once those contours have been exited and their contents
    compiled.  This information is used for outputting debugging info.  */
 
 tree
-build_let (filename, line, vars, body, supercontext, tags)
+build_let (filename, line, vars, subblocks, supercontext, tags)
      char *filename;
      int line;
-     tree vars, body, supercontext, tags;
+     tree vars, subblocks, supercontext, tags;
 {
   register tree t = make_node (LET_STMT);
   STMT_SOURCE_FILE (t) = filename;
   STMT_SOURCE_LINE (t) = line;
   STMT_VARS (t) = vars;
-  STMT_BODY (t) = body;
+  STMT_SUBBLOCKS (t) = subblocks;
   STMT_SUPERCONTEXT (t) = supercontext;
   STMT_BIND_SIZE (t) = 0;
   STMT_TYPE_TAGS (t) = tags;
@@ -1539,63 +1829,6 @@ build_type_variant (type, constp, volatilep)
   current_obstack = ambient_obstack;
   return t;
 }
-
-/* Return a type like TYPE except that its CLASSTYPE_OFFSET
-   is OFFSET.
-
-   Such variant types already made are recorded so that duplicates
-   are not made.
-
-   A variant types should never be used as the type of an expression.
-   Use TYPE_MAIN_VARIANT to find the main variant.  */
-
-tree
-build_classtype_variant (type, offset, virtualp)
-     tree type;
-     tree offset;
-     int virtualp;
-{
-  register tree t, m = CLASSTYPE_MAIN_VARIANT (type);
-  register struct obstack *ambient_obstack = current_obstack;
-  register struct obstack *ambient_saveable_obstack = saveable_obstack;
-  register int lo = TREE_INT_CST_LOW (offset);
-  register int hi = TREE_INT_CST_HIGH (offset);
-  /* First search the chain variants for one that is what we want.  */
-
-  if (hi == 0 && lo == 0)
-    offset = integer_zero_node;
-
-  for (t = m; t; t = CLASSTYPE_NEXT_VARIANT (t))
-    if (virtualp == TREE_VIA_VIRTUAL (t)
-	&& lo == TREE_INT_CST_LOW (CLASSTYPE_OFFSET (t))
-	&& hi == TREE_INT_CST_HIGH (CLASSTYPE_OFFSET (t)))
-      return t;
-
-  /* We need a new one.  */
-  if (TREE_PERMANENT (type))
-    saveable_obstack = &permanent_obstack;
-  current_obstack = saveable_obstack;
-
-  t = copy_node (type);
-  copy_type_lang_specific (t);
-  CLASSTYPE_THIS_VARIANT (t) = t;
-
-  TYPE_POINTER_TO (t) = 0;
-  TYPE_REFERENCE_TO (t) = 0;
-  CLASSTYPE_OFFSET (t) = offset;
-  TREE_VIA_VIRTUAL (t) = virtualp;
-
-  /* Always promise to have TYPE_POINTER_TO filled in.  */
-  build_pointer_type (t);
-
-  /* Add this type to the chain of variants of TYPE.  */
-  CLASSTYPE_NEXT_VARIANT (t) = CLASSTYPE_NEXT_VARIANT (m);
-  CLASSTYPE_NEXT_VARIANT (m) = t;
-
-  current_obstack = ambient_obstack;
-  saveable_obstack = ambient_saveable_obstack;
-  return t;
-}
 
 /* Hashing of types so that we don't make duplicates.
    The entry point is `type_hash_canon'.  */
@@ -1713,6 +1946,10 @@ type_hash_canon (hashcode, type)
       struct obstack *o
 	= TREE_PERMANENT (type) ? &permanent_obstack : saveable_obstack;
       obstack_free (o, type);
+#ifdef GATHER_STATISTICS
+      tree_node_kinds[(int)t_kind]--;
+      tree_node_sizes[(int)t_kind] -= sizeof (struct tree_type);
+#endif
       return t1;
     }
 
@@ -1737,9 +1974,14 @@ type_list_equal (l1, l2)
     {
       if (TREE_VALUE (t1) != TREE_VALUE (t2))
 	return 0;
-      if (TREE_PURPOSE (t1) != TREE_PURPOSE (t2)
-	  && !simple_cst_equal (TREE_PURPOSE (t1), TREE_PURPOSE (t2)))
-	return 0;
+      if (TREE_PURPOSE (t1) != TREE_PURPOSE (t2))
+	{
+	  int cmp = simple_cst_equal (TREE_PURPOSE (t1), TREE_PURPOSE (t2));
+	  if (cmp < 0)
+	    abort ();
+	  if (cmp == 0)
+	    return 0;
+	}
     }
 
   return t1 == t2;
@@ -1786,7 +2028,10 @@ simple_cst_list_equal (l1, l2)
 {
   while (l1 != NULL_TREE && l2 != NULL_TREE)
     {
-      if (! simple_cst_equal (TREE_VALUE (l1), TREE_VALUE (l2)))
+      int cmp = simple_cst_equal (TREE_VALUE (l1), TREE_VALUE (l2));
+      if (cmp < 0)
+	abort ();
+      if (cmp == 0)
 	return 0;
       l1 = TREE_CHAIN (l1);
       l2 = TREE_CHAIN (l2);
@@ -1794,11 +2039,17 @@ simple_cst_list_equal (l1, l2)
   return (l1 == l2);
 }
 
+/* Return truthvalue of whether T1 is the same tree structure as T2.
+   Return 1 if they are the same.
+   Return 0 if they are understandably different.
+   Return -1 if either contains tree structure not understood by
+   this function.  */
 int
 simple_cst_equal (t1, t2)
      tree t1, t2;
 {
   register enum tree_code code1, code2;
+  int cmp;
 
   if (t1 == t2)
     return 1;
@@ -1830,7 +2081,8 @@ simple_cst_equal (t1, t2)
 
     case STRING_CST:
       return TREE_STRING_LENGTH (t1) == TREE_STRING_LENGTH (t2)
-	&& !strcmp (TREE_STRING_POINTER (t1), TREE_STRING_POINTER (t2));
+	&& !bcmp (TREE_STRING_POINTER (t1), TREE_STRING_POINTER (t2),
+		  TREE_STRING_LENGTH (t1));
 
     case CONSTRUCTOR:
       abort ();
@@ -1840,12 +2092,15 @@ simple_cst_equal (t1, t2)
 
     case CALL_EXPR:
     case NEW_EXPR:
-      return (simple_cst_equal (TREE_OPERAND (t1, 0), TREE_OPERAND (t2, 0))
-	      && simple_cst_list_equal (TREE_OPERAND (t1, 1), TREE_OPERAND (t2, 1)));
+      cmp = simple_cst_equal (TREE_OPERAND (t1, 0), TREE_OPERAND (t2, 0));
+      if (cmp <= 0)
+	return cmp;
+      return simple_cst_list_equal (TREE_OPERAND (t1, 1), TREE_OPERAND (t2, 1));
 
     case COMPONENT_REF:
-      return (TREE_OPERAND (t1, 1) == TREE_OPERAND (t2, 1)
-	      && simple_cst_equal (TREE_OPERAND (t1, 0), TREE_OPERAND (t2, 0)));
+      if (TREE_OPERAND (t1, 1) == TREE_OPERAND (t2, 1))
+	return simple_cst_equal (TREE_OPERAND (t1, 0), TREE_OPERAND (t2, 0));
+      return 0;
 
     case VAR_DECL:
     case PARM_DECL:
@@ -1860,8 +2115,10 @@ simple_cst_equal (t1, t2)
     case TRUNC_MOD_EXPR:
     case LSHIFT_EXPR:
     case RSHIFT_EXPR:
-      return (simple_cst_equal (TREE_OPERAND (t1, 0), TREE_OPERAND (t2, 0))
-	      && simple_cst_equal (TREE_OPERAND (t1, 1), TREE_OPERAND (t2, 1)));
+      cmp = simple_cst_equal (TREE_OPERAND (t1, 0), TREE_OPERAND (t2, 0));
+      if (cmp <= 0)
+	return cmp;
+      return simple_cst_equal (TREE_OPERAND (t1, 1), TREE_OPERAND (t2, 1));
 
     case NEGATE_EXPR:
     case ADDR_EXPR:
@@ -1870,7 +2127,7 @@ simple_cst_equal (t1, t2)
       return simple_cst_equal (TREE_OPERAND (t1, 0), TREE_OPERAND (t2, 0));
 
     default:
-      abort ();
+      return -1;
     }
 }
 
@@ -1927,11 +2184,10 @@ build_index_type (maxval)
 {
   register tree itype = make_node (INTEGER_TYPE);
   int maxint = TREE_INT_CST_LOW (maxval);
-  TYPE_PRECISION (itype) = BITS_PER_WORD;
+  TYPE_PRECISION (itype) = TYPE_PRECISION (sizetype);
   TYPE_MIN_VALUE (itype) = build_int_2 (0, 0);
-  TREE_TYPE (TYPE_MIN_VALUE (itype)) = itype;
-  TYPE_MAX_VALUE (itype) = maxval;
-  TREE_TYPE (maxval) = itype;
+  TREE_TYPE (TYPE_MIN_VALUE (itype)) = sizetype;
+  TYPE_MAX_VALUE (itype) = convert (sizetype, maxval);
   TYPE_MODE (itype) = SImode;
   TYPE_SIZE (itype) = TYPE_SIZE (sizetype);
   TYPE_SIZE_UNIT (itype) = TYPE_SIZE_UNIT (sizetype);
@@ -1970,12 +2226,6 @@ build_array_type (elt_type, index_type)
 
   if (TYPE_SIZE (t) == 0)
     layout_type (t);
-
-  /* Push these needs up so that initialization takes place
-     more easily.  */
-  TYPE_NEEDS_CONSTRUCTING (t) = TYPE_NEEDS_CONSTRUCTING (TYPE_MAIN_VARIANT (elt_type));
-  TYPE_NEEDS_DESTRUCTOR (t) = TYPE_NEEDS_DESTRUCTOR (TYPE_MAIN_VARIANT (elt_type));
-
   return t;
 }
 
@@ -2067,7 +2317,7 @@ build_method_type (basetype, type)
   if (TREE_CODE (type) != FUNCTION_TYPE)
     abort ();
 
-  TYPE_METHOD_BASETYPE (t) = basetype;
+  TYPE_METHOD_BASETYPE (t) = TYPE_MAIN_VARIANT (basetype);
   TREE_TYPE (t) = TREE_TYPE (type);
 
   /* The actual arglist for this function includes a "hidden" argument
@@ -2101,7 +2351,7 @@ build_offset_type (basetype, type)
   /* Make a node of the sort we want.  */
   t = make_node (OFFSET_TYPE);
 
-  TYPE_OFFSET_BASETYPE (t) = basetype;
+  TYPE_OFFSET_BASETYPE (t) = TYPE_MAIN_VARIANT (basetype);
   TREE_TYPE (t) = type;
 
   /* If we already have such a type, use the old one and free this one.  */
@@ -2113,6 +2363,33 @@ build_offset_type (basetype, type)
 
   return t;
 }
+
+/* Return the innermost context enclosing FNDECL that is
+   a RECORD_TYPE or UNION_TYPE, or zero if none.
+   TYPE_DECLs and FUNCTION_DECLs are transparent to this function.  */
+
+tree
+decl_type_context (fndecl)
+     tree fndecl;
+{
+  tree context = DECL_CONTEXT (fndecl);
+
+  while (context)
+    {
+      if (TREE_CODE (context) == RECORD_TYPE
+	  || TREE_CODE (context) == UNION_TYPE)
+	return context;
+      if (TREE_CODE (context) == TYPE_DECL
+	  || TREE_CODE (context) == FUNCTION_DECL)
+	context = DECL_CONTEXT (context);
+      else if (TREE_CODE (context) == LET_STMT)
+	context = STMT_SUPERCONTEXT (context);
+      else
+	/* Unhandled CONTEXT!?  */
+	abort ();
+    }
+  return NULL_TREE;
+}
 
 /* Return OP, stripped of any conversions to wider types as much as is safe.
    Converting the value back to OP's type makes a value equivalent to OP.
@@ -2353,9 +2630,29 @@ void
 dump_tree_statistics ()
 {
   int i;
-  fprintf (stderr, "%d tree nodes created\n", tree_node_counter);
+  int total_nodes, total_bytes;
+  extern struct obstack class_obstack;
+
+  fprintf (stderr, "\n%d tree nodes created\n\n", tree_node_counter);
+#ifdef GATHER_STATISTICS
+  fprintf (stderr, "Kind                  Nodes     Bytes\n");
+  fprintf (stderr, "-------------------------------------\n");
+  total_nodes = total_bytes = 0;
   for (i = 0; i < (int) all_kinds; i++)
-    fprintf (stderr, "%s: %d\n", tree_node_kind_names[i], tree_node_kinds[i]);
+    {
+      fprintf (stderr, "%-20s %6d %9d\n", tree_node_kind_names[i],
+	       tree_node_kinds[i], tree_node_sizes[i]);
+      total_nodes += tree_node_kinds[i];
+      total_bytes += tree_node_sizes[i];
+    }
+  fprintf (stderr, "%-20s        %9d\n", "identifier names", id_string_size);
+  fprintf (stderr, "-------------------------------------\n");
+  fprintf (stderr, "%-20s %6d %9d\n", "Total", total_nodes, total_bytes);
+  fprintf (stderr, "-------------------------------------\n");
+#else
+  fprintf (stderr, "(No per-node statistics)\n");
+#endif
+  print_obstack_statistics ("class_obstack", &class_obstack);
   print_obstack_statistics ("permanent_obstack", &permanent_obstack);
   print_obstack_statistics ("maybepermanent_obstack", &maybepermanent_obstack);
   print_search_statistics ();
